
from __future__ import annotations

import abc
import typing as t

from nr.util.generic import T


class Transaction(abc.ABC):
  """
  Represents a transaction object that can be commited or aborted.
  """

  @abc.abstractmethod
  def commit(self) -> None:
    ...

  @abc.abstractmethod
  def abort(self) -> None:
    ...

  def __enter__(self: T) -> 'T':
    return self

  def __exit__(self, exc_v, exc_t, exc_tb) -> None:
    if exc_v is None:
      self.commit()
    else:
      self.abort()


class KeyValueStore(abc.ABC):
  """
  Interface for key-value stores.
  """

  @abc.abstractmethod
  def get(self, key: str) -> bytes:
    ...

  @abc.abstractmethod
  def set(self, key: str, data: bytes, exp: int | None = None) -> None:
    ...

  @abc.abstractmethod
  def delete(self, key: str) -> None:
    ...

  @abc.abstractmethod
  def keys(self, prefix: str = '') -> t.Iterable[str]:
    ...

  @abc.abstractmethod
  def count(self, prefix: str = '') -> int:
    ...




""" ASGI/WSGI application launchers. """

from __future__ import annotations

import abc
import base64
import dataclasses
import pickle
import os

from .._app import App


class AWSGILauncher(abc.ABC):
  """
  Interface for AWSGI/WSGI application launchers.
  """

  @abc.abstractmethod
  def launch(self, entrypoint: str, factory: bool) -> None: ...


try:
  from databind.core.annotations import union  # type: ignore[import]
except ImportError:
  pass
else:
  union(
    union.Subtypes.chain(
      union.Subtypes.dynamic({
        'uvicorn': lambda: __import__(__name__ + '.uvicorn', fromlist=['UvicornLauncher']).UvicornLauncher,
      }),
      union.Subtypes.entrypoint('nr.util.awsgi.launcher'),
    ),
    style=union.Style.flat
  )(AWSGILauncher)


def launch(launcher: AWSGILauncher, app: str | App, factory: bool = False) -> None:
  """
  Launches the application using the specified ASGI/WSGI launcher. If an #App instance is specified, it will be
  pickled in this process and unpickled in subprocesses to avoid limitations of finding an application just by
  an entrypoint.
  """

  if not isinstance(app, str):
    if not isinstance(app, App):
      raise RuntimeError(f'expected str|App, got {type(app).__name__}')

    os.environ['_NR_UTILS_AWSGI_PICKLED_APP'] = base64.b64encode(pickle.dumps(app)).decode('ascii')
    app = __name__ + ':_pickled_app'
    factory = True

  launcher.launch(app, factory)


def _pickled_app():
  """ Internal. Loader for a pickled app. """

  app: App = pickle.loads(base64.b64decode(os.environ.pop('_NR_UTILS_AWSGI_PICKLED_APP').encode('ascii')))
  app.initialize()

  if app.asgi_app:
    return app.asgi_app
  elif app.wsgi_app:
    return app.wsgi_app
  else:
    raise RuntimeError(f'{type(app).__name__} does not provide an asgi_app or wsgi_app')


@dataclasses.dataclass
class BaseAWSGIConfig:

  @dataclasses.dataclass
  class Files:
    pidfile: str | None = None
    stdout: str | None = None
    stderr: str | None = None

  @dataclasses.dataclass
  class Ssl:
    cert: str
    key: str

  host: str = '127.0.0.1'
  port: int = 8000
  daemonize: bool = False
  files: Files = dataclasses.field(default_factory=Files)
  ssl: Ssl | None = None
  num_workers: int | None = None
  additional_options: list[str] = dataclasses.field(default_factory=list)





import io
import typing as t

SubstRange = t.Tuple[int, int, str]


def substitute_ranges(text: str, ranges: t.Iterable[SubstRange], is_sorted: bool = False) -> str:
  """ Replaces parts of *text* using the specified *ranges* and returns the new text. Ranges must not overlap.
  *is_sorted* can be set to `True` if the input *ranges* are already sorted from lowest to highest starting index to
  optimize the function.
  """

  if not is_sorted:
    ranges = sorted(ranges, key=lambda x: x[0])

  out = io.StringIO()
  max_end_index = 0
  for index, (istart, iend, subst) in enumerate(ranges):
    if iend < istart:
      raise ValueError(f'invalid range at index {index}: (istart: {istart!r}, iend: {iend!r})')
    if istart < max_end_index:
      raise ValueError(f'invalid range at index {index}: overlap with previous range')

    subst = str(subst)
    out.write(text[max_end_index:istart])
    out.write(subst)
    max_end_index = iend

  out.write(text[max_end_index:])
  return out.getvalue()




from __future__ import annotations

import dataclasses
import functools
import logging
import os
import typing as t
from importlib import import_module

from nr.util.process import spawn_daemon, replace_stdio, Pidfile
from . import AWSGILauncher, BaseAWSGIConfig

if t.TYPE_CHECKING:
  import flask   # type: ignore

logger = logging.getLogger(__name__)


def load_entrypoint(ep: str) -> t.Any:
  """
  Loads an entrypoint definition of the form `<module>:<member>`.
  """

  module_name, member_name = ep.split(':', 1)
  return getattr(import_module(module_name), member_name)



@dataclasses.dataclass
class FlaskRunner(AWSGILauncher, BaseAWSGIConfig):
  """
  Runs a #flask.Flask application using the Flask development server. This is not suitable
  for production, though very useful during development due to the built-in debugging
  capabilities.

  Note: Daemonizing the Flask runner is currently only available on systems that support
  #os.fork().
  """

  #: Enable the Flask server's debug capabilities.
  debug: bool = False

  #: Automatically defaults to `True` if #debug is enabled.
  use_reloader: bool | None = None

  #: Enable stdio redirects.
  redirect_stdio: bool = False

  def get_status(self) -> Pidfile.Status:
    """
    Returns the status of the application, i.e. whether it is currently running or not. This
    method is only relevant when the application was started in the background.
    """

    filename = self.files.pidfile
    if not filename:
      return Pidfile.Status.UNKNOWN
    return Pidfile(filename).get_status()

  def is_main(self) -> bool:
    """
    Checks whether this is the main process, which can be the reloader thread in case
    the #FlaskRunner is used.
    """

    if os.getenv('_FLASK_USE_RELOAER') == 'true':
      return not os.getenv('WERKZEUG_RUN_MAIN')

    return True

  def launch(self, entrypoint: str, factory: bool) -> None:
    import flask  # type: ignore[module]

    # Set up the environment.
    use_reloader = self.debug if self.use_reloader is None else self.use_reloader
    if use_reloader:
      os.environ['_FLASK_USE_RELOAER'] = 'true'
    if self.debug:
      os.environ['FLASK_DEBUG'] = 'true'

    # Load the Flask application.
    app = load_entrypoint(entrypoint)
    if factory:
      app = app()
    if not isinstance(app, flask.Flask):
      raise RuntimeError('entrypoint ({}) must be a Flask application.'.format(entrypoint))

    # Setup stdio redirects.
    if self.redirect_stdio:
      stdout, stderr = self._init_redirects(self.files)
    else:
      stdout, stderr = None, None

    if self.files.pidfile:
      os.makedirs(os.path.dirname(self.files.pidfile), exist_ok=True)

    if self.ssl:
      ssl_context = (self.ssl.cert, self.ssl.key)
    else:
      ssl_context = None

    run = functools.partial(
      self._run,
      app=app,
      stdout=stdout,
      stderr=stderr,
      pidfile=self.files.pidfile,
      daemon=self.daemonize,
      host=self.host,
      port=self.port,
      debug=self.debug,
      use_reloader=use_reloader,
      ssl_context=ssl_context)

    if self.is_main() and self.get_status() == Pidfile.Status.RUNNING:
      raise RuntimeError('Application is already running.')

    if self.daemonize:
      spawn_daemon(run)
    else:
      run()

  def _init_redirects(self, files: BaseAWSGIConfig.Files) -> tuple[t.TextIO | None, t.TextIO | None]:
    if files.stdout:
      os.makedirs(os.path.dirname(files.stdout), exist_ok=True)
      stdout = open(files.stdout, 'a+')
    else:
      stdout = None

    stderr: t.TextIO | None
    if files.stderr and files.stderr != files.stdout:
      os.makedirs(os.path.dirname(files.stderr), exist_ok=True)
      stderr = open(files.stderr, 'a+')
    elif files.stderr == files.stdout:
      stderr = stdout
    else:
      stderr = None

    return stdout, stderr

  def _run(
    self,
    app: 'flask.Flask',
    stdout: t.TextIO | None,
    stderr: t.TextIO | None,
    pidfile: str | None,
    daemon: bool,
    host: str,
    port: int,
    debug: bool,
    use_reloader: bool,
    ssl_context: tuple[str, str]
  ) -> None:
    """
    Internal function to actually invoke the Flask application after everything was prepared.
    """

    if stdout or stderr:
      replace_stdio(None, stdout, stderr)

    if pidfile:
      with open(pidfile, 'w') as fp:
        fp.write(str(os.getpid()))

    if daemon:
      logger.info('Process %s started.', os.getpid())

    try:
      app.run(
        host=host,
        port=port,
        debug=debug,
        use_reloader=use_reloader,
        ssl_context=ssl_context,
      )
    finally:
      if (not use_reloader or os.getenv('WERKZEUG_RUN_MAIN') == 'true') and pidfile:
        try:
          logger.info('Removing pidfile "%s" from PID %s.', pidfile, os.getpid())
          os.remove(pidfile)
        except OSError as exc:
          logger.exception('Unable to remove "%s".', pidfile)




from __future__ import annotations

from ._types import ASGIApp, WSGIApp


class App:
  """
  Base for ASGI/WSGI applications. The class instance itself is not the ASGI/WSGI app, but instead it is expected to
  provide that object after and initialization procedure. An #App subclass should encapsulate all logic for the
  entirety of the application.
  """

  asgi_app: ASGIApp | None = None
  wsgi_app: WSGIApp | None = None

  def initialize(self) -> None:
    """ Delayed initializer, called after the app is handed to the ASGI/WSGI runner to avoid pickling errors. """




from ._pkg_resources import iter_entrypoints, load_entrypoint, NoSuchEntrypointError
from ._plugin_registry import PluginRegistry

__all__ = [
  'iter_entrypoints',
  'load_entrypoint',
  'NoSuchEntrypointError',
  'PluginRegistry',
]




from __future__ import annotations

import dataclasses
import sys


@dataclasses.dataclass
class Callsite:
  filename: str
  lineno: int
  code_name: str
  name: str | None


def get_callsite(stackdepth: int = 1) -> Callsite:
  """ Returns information on the callsite of the specified stackdepth. Defaults to 1 to point to the immediate
  stack frame above from where the function is called. The returned object contains information on the filename
  and line number. """

  frame = sys._getframe(stackdepth + 1)
  try:
    return Callsite(frame.f_code.co_filename, frame.f_lineno, frame.f_code.co_name, frame.f_globals.get('__name__'))
  finally:
    del frame




from __future__ import annotations
import typing as t
import typing_extensions as te

from nr.util.generic import T

_Message: te.TypeAlias = 'str | t.Callable[[], str]'


def _get_message(message: _Message) -> str:
  if isinstance(message, str):
    return message
  else:
    return message()


def assure(v: T | None, msg: _Message | None = None) -> T:
  """ Assures that *v* is not `None` and returns it. If the value is in fact `None`, a {@link ValueError} will
  be raised raised. """

  if v is None:
    raise ValueError(_get_message(msg) if msg else 'expected value to not be None')
  return v




""" Helpers to implement a plugin infrastructure in Python. """

from __future__ import annotations

import logging
import pkg_resources
import typing as t
import typing_extensions as te

from nr.util.generic import T

logger = logging.getLogger(__name__)

# NOTE (@NiklasRosenstein): I wished we could use a TypeVar bound to a protocol instead of T, but mypy does
#   not seem to like it.


class NoSuchEntrypointError(RuntimeError):
  pass


@t.overload
def load_entrypoint(group: str, name: str) -> t.Any: ...


@t.overload
def load_entrypoint(group: type[T], name: str) -> type[T]: ...


def load_entrypoint(group: str | type[T], name: str) -> t.Any | type[T]:
  """ Load a single entrypoint value. Raises a #RuntimeError if no such entrypoint exists. """

  if isinstance(group, type):
    group_name = group.ENTRYPOINT  # type: ignore
  else:
    group_name = group

  for ep in pkg_resources.iter_entry_points(group_name, name):
    value = ep.load()
    break
  else:
    raise NoSuchEntrypointError(f'no entrypoint "{name}" in group "{group}"')

  if isinstance(group, type):
    if not isinstance(value, type):
      raise TypeError(f'entrypoint "{name}" in group "{group}" is not a type (found "{type(value).__name__}")')
    if not issubclass(value, group):  # type: ignore
      raise TypeError(f'entrypoint "{name}" in group "{group}" is not a subclass of {group.__name__}')

  return value


_Iter_Entrypoints_1: te.TypeAlias = 't.Iterator[pkg_resources.EntryPoint]'
_Iter_Entrypoints_2: te.TypeAlias = 't.Iterator[tuple[str, t.Callable[[], type[T]]]]'


@t.overload
def iter_entrypoints(group: str) -> _Iter_Entrypoints_1: ...


@t.overload
def iter_entrypoints(group: type[T]) -> _Iter_Entrypoints_2: ...


def iter_entrypoints(group: str | type[T]) -> _Iter_Entrypoints_1 | _Iter_Entrypoints_2:
  """ Loads all entrypoints from the given group. """

  if isinstance(group, type):
    group_name = group.ENTRYPOINT  # type: ignore
  else:
    group_name = group

  def _make_loader(ep: pkg_resources.EntryPoint) -> t.Callable[[], type[T]]:
    def loader():
      assert isinstance(group, type)
      value = ep.load()
      if not isinstance(value, type):
        raise TypeError(f'entrypoint "{ep.name}" in group "{group}" is not a type (found "{type(value).__name__}")')
      if not issubclass(value, group):  # type: ignore
        raise TypeError(f'entrypoint "{ep.name}" in group "{group}" is not a subclass of {group.__name__}')
      return value
    return loader

  for ep in pkg_resources.iter_entry_points(group_name):
    if isinstance(group, type):
      yield ep.name, _make_loader(ep)
    else:
      yield ep




import enum


class Attribute(enum.Enum):
  RESET = 0
  BOLD = 1
  FAINT = 2
  ITALIC = 3
  UNDERLINE = 4
  SLOW_BLINK = 5
  RAPID_BLINK = 6
  REVERSE_VIDEO = 7
  CONCEAL = 8
  CROSSED_OUT = 9
  FONT_0 = 10
  FONT_1 = 11
  FONT_2 = 12
  FONT_3 = 13
  FONT_4 = 14
  FONT_5 = 15
  FONT_6 = 16
  FONT_7 = 17
  FONT_8 = 18
  FONT_9 = 19
  FRAKTUR = 20
  DOUBLY_UNDERLINE = 21
  NORMAL_INTENSITY = 22
  NOT_ITALIC = 23
  UNDERLINE_OFF = 24
  BLINK_OFF = 25
  REVERSE_OFF = 27
  REVEAL = 28
  CROSSED_OUT_OFF = 29
  FRAMED = 51
  ENCIRCLED = 52
  OVERLINED = 53
  FRAMED_OFF = 54
  OVERLINED_OFF = 55




from __future__ import annotations
import contextlib
import os
import tempfile
import typing as t
import typing_extensions as te
from pathlib import Path

StrPath: te.TypeAlias = 'str | Path'


@t.overload
def atomic_write(
  path: StrPath,
  mode: te.Literal['w'],
  rename_mode: te.Literal['posix', 'windows'] | None,
) -> t.ContextManager[t.TextIO]:
  ...


@t.overload
def atomic_write(
  path: StrPath,
  mode: te.Literal['wb'],
  rename_mode: te.Literal['posix', 'windows'] | None,
) -> t.ContextManager[t.BinaryIO]:
  ...


@contextlib.contextmanager  # type: ignore
def atomic_write(
  path: StrPath,
  mode: te.Literal['w', 'wb'],
  rename_mode: te.Literal['posix', 'windows'] | None = None,
) -> t.Iterator[t.IO]:
  """ Write to a temporarily file, then rename on file closure. If an error occurs while the context manager is active,
  the temporary file will be deleted instead and if an original file existed before it will not be modified. On
  Windows systems, the file cannot be replaced in an atomic operation, so it will need to be deleted first. """

  if rename_mode is None:
    if os.name == 'nt':
      rename_mode = 'windows'
    else:
      rename_mode = 'posix'

  with tempfile.NamedTemporaryFile(mode, delete=False) as fp:
    try:
      yield fp
    except:
      os.remove(fp.name)
      raise
    else:
      fp.flush()
      os.fsync(fp.fileno())
      if rename_mode == 'windows' and os.path.isfile(path):
        os.remove(path)
      os.rename(fp.name, path)


@t.overload
def atomic_swap(
  path: StrPath,
  mode: te.Literal['w'],
  always_revert: bool,
) -> t.ContextManager[t.TextIO]:
  ...


@t.overload
def atomic_swap(
  path: StrPath,
  mode: te.Literal['wb'],
  always_revert: bool,
) -> t.ContextManager[t.BinaryIO]:
  ...


@contextlib.contextmanager  # type: ignore
def atomic_swap(
  path: StrPath,
  mode: te.Literal['w', 'wb'],
  always_revert: bool = False,
) -> t.Iterator[t.IO]:
  """ Similar to #atomic_write(), only that this function writes to the *path* directlty instead of a temporary file,
  and save the original version of the file next to it in the same directory by temporarily renaming it. If the context
  exits without error, the old file will be removed. Otherwise, the new file will be deleted and the old file will be
  renamed back to *path*. If *always_revert* is enabled, the original file will be restored even if the context exits
  without errors. """

  path = Path(path)

  with tempfile.NamedTemporaryFile(mode, prefix=path.stem + '~', suffix='~' + path.suffix, dir=path.parent) as old:
    old.close()
    os.rename(path, old.name)

    def _revert():
      if path.is_file():
        path.unlink()
      os.rename(old.name, path)

    try:
      with path.open(mode) as new:
        yield new
    except:
      _revert()
      raise
    else:
      if always_revert:
        _revert()
      else:
        os.remove(old.name)




from __future__ import annotations

import typing as t
from deprecated import deprecated
from nr.util.generic import T


class SequenceWalker(t.Generic[T]):
  """ The sequence walker is a utility class to walk over the items of a sequence in an easy to use manner. """

  def __init__(self, sequence: t.Sequence[T], start_index: int = 0) -> None:
    self.sequence = sequence
    self.index = start_index

  def __bool__(self) -> bool:
    """ Returns `True` if the walker is pointing to an element in the sequence, `False` if not. """
    return self.index < len(self.sequence)

  @property
  def current(self) -> T:
    """ Returns the item in the sequence that the walker is currently pointing to. """
    return self.sequence[self.index]

  def has_next(self) -> bool:
    """ Returns `True` if calling {@meth next()} will be successful, i.e. if there is a next element that the walker
    can move to. """
    return self.index < len(self.sequence) - 1

  def next(self) -> T:
    """ Returns the next element in the sequence. Note that this cannot be used to advance the walker beyond the
    end of the sequence; use #advance() instead. """
    if self.has_next():
      self.index += 1
      return self.sequence[self.index]
    else:
      raise ValueError('no next element for the walker to move to')

  def advance(self) -> T | None:
    if self:
      self.index += 1
    if self.index < len(self.sequence):
      return self.sequence[self.index]
    return None

  def safe_iter(self, strict_forward: bool = True) -> t.Iterator[T]:
    """ A utility iterator that yields the current line of the scanner at each step of the iteration. After each
    step, it ensures that the Scanner was at least moved forward by one index, or modified at all if *strict* is
    disabled. """

    while self:
      index = self.index
      yield self.current
      if self.index == index:
        raise RuntimeError('scanner was not advanced')
      if strict_forward and self.index < index:
        raise RuntimeError('scanner as moved backwards but strict_forward is enabled')



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['util',
 'util.algorithm',
 'util.annotations',
 'util.atomic',
 'util.awsgi',
 'util.awsgi.launcher',
 'util.config',
 'util.digraph',
 'util.digraph.algorithm',
 'util.fs',
 'util.functional',
 'util.generic',
 'util.git',
 'util.inspect',
 'util.io',
 'util.iter',
 'util.keyvalue',
 'util.logging.filters',
 'util.logging.formatters',
 'util.parsing',
 'util.parsing._tokenizer',
 'util.plugins',
 'util.process',
 'util.proxy',
 'util.safearg',
 'util.singleton',
 'util.task',
 'util.terminal.colors',
 'util.text',
 'util.weak']

package_data = \
{'': ['*']}

install_requires = \
['deprecated>=1.2.0,<2.0.0', 'typing-extensions>=3.0.0']

setup_kwargs = {
    'name': 'nr.util',
    'version': '0.7.6',
    'description': 'General purpose Python utility library.',
    'long_description': '# nr.util\n\nGeneral purpose Python utility library.\n\nCompatible with Python 3.7 and higher.\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/NiklasRosenstein/python-nr.util',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)




from ._digraph import DiGraph, UnknownEdgeError, UnknownNodeError, E, K, N

__all__ = [
  'remove_with_predecessors',
  'topological_sort',
  'DiGraph',
  'UnknownEdgeError',
  'UnknownNodeError',
]




from __future__ import annotations

import dataclasses
import re
import typing as t

from ._attribute import Attribute
from ._color import Color, parse_color


@dataclasses.dataclass
class Style:
  """ A style is a combination of foreground and background color, as well as a list of attributes. """

  RESET: t.ClassVar[Style]

  fg: Color | None = None
  bg: Color | None = None
  attrs: list[Attribute] | None = None

  def __init__(
    self,
    fg: Color | str | None = None,
    bg: Color | str | None = None,
    attrs: t.Sequence[Attribute | str] | str | None = None,
  ) -> None:
    """ The constructor allows you to specify all arguments also as strings. The foreground and background are parsed
    with #parse_color(). The *attrs* can be a comma-separated string. """

    if isinstance(fg, str):
      fg = parse_color(fg)
    if isinstance(bg, str):
      bg = parse_color(bg)
    if isinstance(attrs, str):
      attrs = [x.strip() for x in attrs.split(',') if x.strip()]
    self.fg = fg
    self.bg = bg
    if attrs is None:
      self.attrs = None
    else:
      self.attrs = []
      for attr in attrs:
        if isinstance(attr, str):
          self.attrs.append(Attribute[attr.upper()])
        else:
          self.attrs.append(attr)

  def to_escape_sequence(self) -> str:
    seq = []
    if self.fg:
      seq.append(self.fg.as_foreground())
    if self.bg:
      seq.append(self.bg.as_background())
    seq.extend(str(attr.value) for attr in self.attrs or ())
    return '\033[' + ';'.join(seq) + 'm'


Style.RESET = Style(attrs='reset')


class StyleManager:
  """ Allows you to register styles and format text using HTML-style tags. """

  TAG_EXPR = r'<([^>=]+)([^>]*)>(.*)</\1>'

  def __init__(self) -> None:
    self._styles: dict[str, Style] = {}

  def add_style(
    self,
    name: str,
    fg: Color | str | None = None,
    bg: Color | str | None = None,
    attrs: list[Attribute | str] | str | None = None,
  ) -> None:
    self._styles[name] = Style(fg, bg, attrs)

  def parse_style(self, style_string: str, safe: bool = False) -> Style:
    """ Parses a style string that is valid inside an opening HTML-style tag accepted in strings by #format(). """

    parts = style_string.split(';')
    style: Style = Style()
    for part in parts:
      try:
        if part.startswith('fg='):
          style = Style(parse_color(part[3:]), style.bg, style.attrs)
        elif part.startswith('bg='):
          style = Style(style.fg, parse_color(part[3:]), style.attrs)
        elif part.startswith('attr='):
          style = Style(style.fg, style.bg, (style.attrs or []) + [Attribute[part[5:].upper()]])
        else:
          style = self._styles[part]
      except (ValueError, KeyError):
        if not safe:
          raise

    return style

  def format(self, text: str, safe: bool = False, repl: t.Callable[[str, str], str] | None = None) -> str:
    """ Formats text that contains HTML-style tags that represent styles in the style manager. In addition, special
    tags `<fg={color}>`, `<bg={color}>` or <attr={attrs}>` can be used to manually specify the exact styling of the
    text and the can be combined (such as `<bg=bright red;attr=underline>`). If *safe* is set to `True`, tags
    referencing styles that are unknown to the manager are ignored. """

    def _regex_sub(m: re.Match) -> str:
      style_string = m.group(1) + m.group(2)
      content = m.group(3)
      if repl is None:
        style = self.parse_style(style_string, safe)
        return style.to_escape_sequence() + content + Style.RESET.to_escape_sequence()
      else:
        return repl(style_string, content)

    upper_limit = 15
    prev_text = text
    for _ in range(upper_limit):
      text = re.sub(self.TAG_EXPR, _regex_sub, text, flags=re.S | re.M)
      if prev_text == text:
        break
      prev_text = text

    return prev_text

  @classmethod
  def strip_tags(cls, text: str) -> str:
    return cls().format(text, True, lambda _, s: s)




import functools
import operator
import os
import stat
import typing as t
from pathlib import Path


def flags(flags: int, modstring: str) -> int:
  """ Modifies the stat flags according to *modstring*, mirroring the syntax for POSIX `chmod`. """

  mapping = {
    'r': (stat.S_IRUSR, stat.S_IRGRP, stat.S_IROTH),
    'w': (stat.S_IWUSR, stat.S_IWGRP, stat.S_IWOTH),
    'x': (stat.S_IXUSR, stat.S_IXGRP, stat.S_IXOTH)
  }

  target, direction = 'a', None
  for c in modstring:
    if c in '+-':
      direction = c
      continue
    if c in 'ugoa':
      target = c
      direction = None  # Need a - or + after group specifier.
      continue
    if c in 'rwx' and direction and direction in '+-':
      if target == 'a':
        mask = functools.reduce(operator.or_, mapping[c])
      else:
        mask = mapping[c]['ugo'.index(target)]
      if direction == '-':
        flags &= ~mask
      else:
        flags |= mask
      continue
    raise ValueError('invalid chmod: {!r}'.format(modstring))

  return flags


def repr(flags: int) -> str:
  """ Returns a string representation of the access flags *flags*. """

  template = 'rwxrwxrwx'
  order = (stat.S_IRUSR, stat.S_IWUSR, stat.S_IXUSR,
           stat.S_IRGRP, stat.S_IWGRP, stat.S_IXGRP,
           stat.S_IROTH, stat.S_IWOTH, stat.S_IXOTH)
  return ''.join(template[i] if flags&x else '-' for i, x in enumerate(order))


def update(path: t.Union[str, Path], modstring: str) -> None:
  """ Updates the mode of the *path* in a `chmod` like fashion. """

  os.chmod(path, flags(os.stat(path).st_mode, modstring))




import typing as t
from pathlib import Path


def recurse_directory(path: Path) -> t.Iterator[Path]:
  for item in path.iterdir():
    yield item
    if item.is_dir():
      yield from recurse_directory(item)


def walk_up(path: Path) -> t.Iterator[Path]:
  prev = None
  while path != prev:
    yield path
    prev, path = path, path.parent




""" Provides a logging formatter that understands color hints in the message and decorates it with """

from __future__ import annotations

import logging
import typing_extensions as te

from nr.util.terminal.colors import StyleManager
from nr.util.singleton import NotSet


def get_default_styles() -> StyleManager:
  manager = StyleManager()
  manager.add_style('info', 'blue')
  manager.add_style('warning', 'magenta')
  manager.add_style('error', 'red')
  manager.add_style('critical', 'bright red', None, 'bold,underline')
  return manager


class TerminalColorFormatter(logging.Formatter):
  """ A formatter that enhances text decorated with HTML-style tags with ANSI terminal colors. It can also be
  configured to eliminate the HTML tags instead of converting them to terminal styles. """

  def __init__(self, fmt: str, styles: StyleManager | None | NotSet = NotSet.Value) -> None:
    super().__init__(fmt)
    self.styles = get_default_styles() if styles is NotSet.Value else styles

  def format(self, record: logging.LogRecord) -> str:
    message = super().format(record)
    if self.styles is None:
      return StyleManager.strip_tags(message)
    else:
      return self.styles.format(message, True)

  def install(self, target: te.Literal['tty', 'notty'] | None = None) -> None:
    """ Install the formatter on stream handlers on all handlers of the root logger that are attached to a TTY,
    or otherwise on all that are not attached to a TTY based on the *target* value. If no value is specified, it
    will install into TTY-attached stream handlers if #styles is set. """

    if target is None:
      target = 'notty' if self.styles is None else 'tty'

    for handler in logging.root.handlers:
      if isinstance(handler, logging.StreamHandler) and handler.stream.isatty():
        if target == 'tty':
          handler.setFormatter(self)
      elif target == 'notty':
          handler.setFormatter(self)
